/////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2006 MySQL AB
//
// The MySQL Connector/ODBC is licensed under the terms of the GPLv2
// <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>, like most
// MySQL Connectors. There are special exceptions to the terms and
// conditions of the GPLv2 as it is applied to this software, see the
// FLOSS License Exception
// <http://www.mysql.com/about/legal/licensing/foss-exception.html>.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published
// by the Free Software Foundation; version 2 of the License.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA

//{{NO_DEPENDENCIES}}
// 
// 99% not needed - phase file out or purge uneeded stuff
// 
// Used by myodbc3.rc
//
#define IDOPTIONS			3
#define DRIVERCONNECT			104
#define MYODBCDLG			163
#define MYOPTIONDLG			164
#define MYSQL_LOGO			165
#define IDC_DSNAME			400
#define IDC_DSNAMETEXT			401
#define IDC_SERVER			402
#define IDC_PASSWORDTEXT		403
#define IDC_DESC			404
#define IDC_USER			405
#define IDC_USERTEXT			406
#define IDC_PASSWORD			407
#define IDC_SERVERTEXT			408
#define IDC_DESCRIPTION			409
#define IDS_MSGTITLE			500
#define IDS_BADDSN			501
#define IDC_PORT			1000
#define IDC_STMT			1001
#define IDC_DB				1005
#define IDC_FLAGTEXT			1007
#define IDC_FLAG			1009
#define IDC_PORTTEXT			1010
#define IDC_CHECK1			1024
#define IDC_CHECK2			1025
#define IDC_CHECK3			1026
#define IDC_CHECK4			1027
#define IDC_CHECK5			1028
#define IDC_CHECK6			1029
#define IDC_CHECK7			1030
#define IDC_CHECK8			1031
#define IDC_CHECK9			1032
#define IDC_CHECK10			1033
#define IDC_CHECK11			1034
#define IDC_CHECK12			1035
#define IDC_CHECK13			1036
#define IDC_CHECK14			1037
#define IDC_CHECK15			1038
#define IDC_CHECK16			1039
#define IDC_CHECK17			1040
#define IDC_CHECK18			1041
#define IDC_CHECK19			1042
#define IDC_CHECK20			1043
#define IDC_CHECK21			1044
#define IDC_CHECK22			1045
#define IDTEST				1063
#define IDC_LOGO			1090
#define IDBACK				1100
#define IDC_STATIC			-1

// Next default values for new objects
//
#if defined (APSTUDIO_INVOKED) && !defined(APSTUDIO_READONLY_SYMBOLS)
#define _APS_NEXT_RESOURCE_VALUE	167
#define _APS_NEXT_COMMAND_VALUE		40013
#define _APS_NEXT_CONTROL_VALUE		1107
#define _APS_NEXT_SYMED_VALUE		101
#endif /* defined (APSTUDIO_INVOKED) && !defined(APSTUDIO_READONLY_SYMBOLS) */
